# Makefile for a directory with several sub-directories 
# Last edited on 2010-03-06 19:33:54 by stolfilocal

# Client may define ${SUBDIRS}, a list of sub-directories
# of the current directory.  "make XXX" will perform "make XXX"
# on each subdirectory where "make" needs to  run.  Alternatively list of 
# subdirectories with specific contents:
#
#   ${LIB_DIRS}       libraries
#   ${PROG_DIRS}      excutable prog packages
#   ${PROG_TEST_DIRS} test runs of program packages
#   ${LIB_TEST_DIRS}  test programs of libraries, and runs thereof
#  
# some actions may be propagated only to some kinds of sub-directories

ifeq "/${SUBDIRS}" "/"
SUBDIRS := ${LIB_DIRS} ${PROG_DIRS} ${LIB_TEST_DIRS} ${PROG_TEST_DIRS}
endif

.PHONY:: \
  default-action \
  all actions \
  clean check \
  setup 

# ----------------------------------------------------------------------
# "make" with no action does "make all" which must be defined by user:
default-action: all

# ----------------------------------------------------------------------
# "make all"
# "make clean"
# "make build"
# "make check"
# These actions are just performed on all sub-directories in the 
# client-defined list ${SUBDIRS}, in order.

all:
	@for dir in ${SUBDIRS}; do \
          ( echo '= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = ='; \
            echo "$$dir"; \
            cd $$dir/. && \
            make all ; \
          ) ; \
        done

build:
	@for dir in ${SUBDIRS}; do \
          ( echo '= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = ='; \
            echo "$$dir"; \
            cd $$dir/. && \
            make build ; \
          ) ; \
        done

check: 
	@for dir in ${SUBDIRS}; do \
          ( echo '= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = ='; \
            echo "$$dir"; \
            cd $$dir/. && \
            make check ; \
          ) ; \
        done

clean: 
	@for dir in ${SUBDIRS}; do \
          ( echo '= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = ='; \
            echo "$$dir"; \
            cd $$dir/. && \
            make clean ; \
          ) ; \
        done

# ----------------------------------------------------------------------
# "make actions" makes the list ${ACTIONS} defined by the user in each
# directory in turn:

actions: 
	for dir in ${SUBDIRS}; do \
          ( echo '= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = ='; \
            echo "$$dir"; \
            cd $$dir/. && \
            make ACTIONS="${ACTIONS}" actions ; \
          ) ; \
        done

# ----------------------------------------------------------------------
# "make setup" creates the user's global repositories for "make install"

