/* See pst_height_map.h */
/* Last edited on 2010-05-04 03:38:52 by stolfi */

#include <assert.h>
#include <affirm.h>
#include <math.h>

#include <pst_height_map.h>
#include <float_image.h>
#include <float_image_mscale.h>
#include <teste_func.h>

float_image_t *pst_height_map_expand_old(float_image_t *JZ, int NXI, int NYI, int expOrder)
  { 
    /* !!! Generalize and move to {float_image_mscale.h} !!! */
    
    demand(expOrder == 2, "{expOrder} other than 2 is not supported yet");
    
    assert(JZ->sz[0] == 1);
    int NXJ = JZ->sz[1]; assert(NXI/2 + 1 == NXJ);
    int NYJ = JZ->sz[2]; assert(NYI/2 + 1 == NYJ);
    float_image_t *IZ = float_image_new(1, NXI, NYI);
    int xI, yI;
    for(yI = 0; yI < NYI; yI++)
      { for(xI = 0; xI < NXI; xI++)
          { /* Get relevant samples from original image: */
            int xJ0 = xI/2, xJ1 = (xI+1)/2;
            int yJ0 = yI/2, yJ1 = (yI+1)/2;
            assert(xJ1 < NXJ); assert(yJ1 < NYJ);
            
	    double d00 = float_image_get_sample(JZ, 0, xJ0, yJ0);
            double d01 = (yJ1 == yJ0 ? d00 : float_image_get_sample(JZ, 0, xJ0, yJ1));
            double d10 = (xJ1 == xJ0 ? d00 : float_image_get_sample(JZ, 0, xJ1, yJ0));
            double d11 = ((xJ1 == xJ0) && (yJ1 == yJ0) ? d00 : float_image_get_sample(JZ, 0, xJ1, yJ1));
            /* Average samples, and scale {Z} up by 2: */
            float v = (d00 + d01 + d10 + d11)/2;
            /* Store sample in expanded image: */
            float_image_set_sample(IZ, 0, xI, yI, v);
          }
      }
    return IZ;
  }

float_image_t *pst_height_map_expand(float_image_t *JZ,float_image_t *JW, int NXI, int NYI, int expOrder)
  { 
     /* !!! Generalize and move to {float_image_mscale.h} !!! */
    
    demand(expOrder == 2, "{expOrder} other than 2 is not supported yet");
    
    assert(JZ->sz[0] == 1);
    int NXJ = JZ->sz[1]; assert(NXI/2 + 1 == NXJ);
    int NYJ = JZ->sz[2]; assert(NYI/2 + 1 == NYJ);
    float_image_t *IZ = float_image_new(1, NXI, NYI);
    int xI, yI;
    for(yI = 0; yI < NYI; yI++)
      { for(xI = 0; xI < NXI; xI++)
          { 
	    int xJ = xI/2;
	    int yJ = yI/2;
	    double v,w;
	    
            if( (xI%2) == 0 )
	    {
	      if( (yI%2) == 0 )
	      {
		v = float_image_get_sample(JZ,0,xJ,yJ);
		w = (JW == NULL ? 1 : float_image_get_sample(JZ,0,xJ,yJ));
	      }
	      else
	      {
		pst_interpolate_four_samples(JZ,JW,0,xJ,yJ-1,xJ,yJ,xJ,yJ+1,xJ,yJ+2,&v,&w);
     	      }
	    }
	    else
	    {
	      if( (yI%2) == 0 )
	      { pst_interpolate_four_samples(JZ,JW,0,xJ-1,yJ,xJ,yJ,xJ+1,yJ,xJ+2,yJ,&v,&w); }
	      else
	      { double wa,wb;
		double va,vb;
		pst_interpolate_four_samples(JZ,JW,0,xJ-1,yJ-1,xJ,yJ,xJ+1,yJ+1,xJ+2,yJ+2,&va,&wa);
		pst_interpolate_four_samples(JZ,JW,0,xJ-1,yJ+2,xJ,yJ+1,xJ+1,yJ,xJ+2,yJ-1,&vb,&wb);
		if( (wa == 0) && (wb == 0) )
		{ w = 0; v = (va+vb)/2.0;}
		else{ w = wa+wb; v = (wa*va + wb*vb)/w;}
	      }
	    }
            /* Dont forget to multiply the interpolation by two ! */
            float_image_set_sample(IZ, 0, xI, yI, (float)2*v);
         }
      }
    return IZ;
  }


float_image_t *pst_height_map_shrink(float_image_t *IZ)
  {
    int NX_IZ = IZ->sz[1];
    int NY_IZ = IZ->sz[2];
    int NX_JZ = IZ->sz[1]/2+1;
    int NY_JZ = IZ->sz[2]/2+1;
 
    float_image_t* JZ = float_image_new(1,NX_JZ,NY_JZ);
    
    int jx,jy;
    for(jy = 0; jy < NY_JZ; jy++)
    {
      for(jx = 0; jx < NX_JZ; jx++)
      {
	int ix = 2*jx, iy = 2*jy;
	double wmm = 0,wmo = 0, wmp = 0, wom = 0, woo = 1, wop = 0, wpm = 0, wpo = 0, wpp = 0;
	double vmm = 0,vmo = 0, vmp = 0, vom = 0,          vop = 0, vpm = 0, vpo = 0, vpp = 0;
	double voo = float_image_get_sample(IZ,0,ix,iy);
	if( (ix > 0) && (ix < (NX_IZ-1) ) )
	{
	  wmo = wpo = 0.5;
	  vmo = float_image_get_sample(IZ,0,ix-1,iy);
	  vpo = float_image_get_sample(IZ,0,ix+1,iy);
	}
	
	if((iy > 0) && (iy < (NY_IZ-1) ))
	{
	  wom = wop = 0.5;
	  vom = float_image_get_sample(IZ,0,ix,iy-1);
	  vop = float_image_get_sample(IZ,0,ix,iy+1);
	}
	
	if( (ix > 0) && (ix < (NX_IZ-1) ) && (iy > 0) && (iy < (NY_IZ-1) ) )
	{
	  wmm = wmp = wpm = wpp = 0.25;
	  vmm = float_image_get_sample(IZ,0,ix-1,iy-1);
	  vmp = float_image_get_sample(IZ,0,ix-1,iy+1);
	  vpm = float_image_get_sample(IZ,0,ix+1,iy-1);
	  vpp = float_image_get_sample(IZ,0,ix+1,iy+1);
	  
	}
	
	double sW = wmm + wmo + wmp + wom + woo + wop + wpm + wpo + wpp;
	double svW = wmm*vmm + wmo*vmo + wmp*vmp + wom*vom + woo*voo + wop*vop + wpm*vpm + wpo*vpo + wpp*vpp;
	
	float_image_set_sample(JZ,0,jx,jy,0.5*svW/sW);
	
      }
    }
  
    
    return JZ;
  }

float_image_t *pst_height_map_compare
  ( float_image_t *AZ,
    float_image_t *BZ,
    float_image_t *W,
    bool_t zero_mean,
    double *sAZP,
    double *sBZP,
    double *sEZP,
    double *sreP
  )
  { 
    assert(AZ->sz[0] == 1);
    assert(BZ->sz[0] == 1);
    int NX = AZ->sz[1]; assert(BZ->sz[1] == NX);
    int NY = AZ->sz[2]; assert(BZ->sz[2] == NY);
    
    if (W != NULL)
      { assert(W->sz[0] == 1);
        assert(W->sz[1] == NX);
        assert(W->sz[2] == NY);
      }
      
    /* Compute the mean values of {AZ,BZ,EZ}: */
    double sum_AZW = 0;
    double sum_BZW = 0;
    double sum_EZW = 0;
    double sum_W = 0;
    int x, y;
    for(y = 0; y < NY; y++)
      { for(x = 0; x < NX; x++)
          { /* Get relevant samples from original image: */
            double vA = float_image_get_sample(AZ, 0, x, y);
            double vB = float_image_get_sample(BZ, 0, x, y);
            double vW = 1.0;
	    if(W != NULL) vW = float_image_get_sample(W, 0, x, y);
            double vE = vA - vB;
            /* Accumulate values: */
            sum_AZW += vA*vW;
            sum_BZW += vB*vW;
            sum_EZW += vE*vW;
            sum_W += vW;
          }
      }
    double avgA = (sum_W == 0 ? 0.5 : sum_AZW/sum_W);
    double avgB = (sum_W == 0 ? 0.5 : sum_BZW/sum_W);
    double avgE = (sum_W == 0 ? 0.0 : sum_EZW/sum_W);
    
    /* Unweighted mean of weights: */
    double avgW = sum_W/(NX*NY);
      
    /* Fill {EZ} and RMS values: */
    float_image_t *EZ = float_image_new(1, NX, NY);
    double sum_AdZ2W = 0.0;
    double sum_BdZ2W = 0.0;
    double sum_EZ2W = 0.0;
    for(y = 0; y < NY; y++)
      { for(x = 0; x < NX; x++)
          { /* Get relevant samples from original image: */
            double vA = float_image_get_sample(AZ, 0, x, y);
            double vB = float_image_get_sample(BZ, 0, x, y);
            double vW = 1.0;
	    if(W != NULL) vW = float_image_get_sample(W, 0, x, y);
            double vE = vA - vB;
            if (zero_mean) { vE = vE - avgE; }
            /* Store difference in error image: */
            float_image_set_sample(EZ, 0, x, y, (float)(sqrt(vW/avgW)*vE));
            /* Accumuate squares: */
            vA = vA - avgA;
            vB = vB - avgB;
            sum_AdZ2W += vW*vA*vA;
            sum_BdZ2W += vW*vB*vB;
            sum_EZ2W += vW*vE*vE;
          }
      }
    /* Compute the RMS values and errors: */
    double sAZ = (sum_W == 0 ? 0.0 : sqrt(sum_AdZ2W/sum_W));
    double sBZ = (sum_W == 0 ? 0.0 : sqrt(sum_BdZ2W/sum_W));
    double sEZ = (sum_W == 0 ? 0.0 : sqrt(sum_EZ2W/sum_W));
    double sMZ = hypot(sAZ,sBZ)/M_SQRT2;
    double sre = (sMZ == 0 ? 0.0 : sEZ/sMZ);
    
    /* Return results: */
    if (sAZP != NULL) { (*sAZP) = sAZ; }
    if (sBZP != NULL) { (*sBZP) = sBZ; }
    if (sEZP != NULL) { (*sEZP) = sEZ; }
    if (sreP != NULL) { (*sreP) = sre; }

    return EZ;
  }

void pst_height_map_level_analyze_and_write
  ( char *filePrefix,
    int level,
    bool_t levelTag,
    int iter,
    bool_t iterTag,
    double change,
    float_image_t *CZ,
    float_image_t *CW,
    float_image_t *RZ,
    float_image_t *EW, 
    bool_t writeImages,
    bool_t writeError,
    int indent
  )
  {
    demand(CZ->sz[0] == 1, "bad CZ channels");
    int NX = CZ->sz[1]; 
    int NY = CZ->sz[2]; 
    
    if (RZ != 0)
      { demand(RZ->sz[0] == 1, "bad RZ channels");
        demand(NX == RZ->sz[1], "bad RZ cols"); 
        demand(NY == RZ->sz[2], "bad RZ rows"); 
      }
    
    if (EW != 0)
      { demand(EW->sz[0] == 1, "bad EW channels");
        demand(NX == EW->sz[1], "bad EW cols"); 
        demand(NY == EW->sz[2], "bad EW rows"); 
      }
      
    if (CW != 0)
      { demand(CW->sz[0] == 1, "bad CW channels");
        demand(NX == CW->sz[1], "bad CW cols"); 
        demand(NY == CW->sz[2], "bad CW rows"); 
      }
      
    int levelQ = (levelTag ? level : -1);
    int iterQ = (iterTag ? iter : -1);
    
    if (writeImages)
      { float_image_mscale_write_file(CZ, filePrefix, levelQ, iterQ, "oZ", indent);
        if(EW != NULL) float_image_mscale_write_file(EW, filePrefix, levelQ, iterQ, "eW", indent);
	if(CW != NULL) float_image_mscale_write_file(CW, filePrefix, levelQ, iterQ, "oW", indent);
      }
      
    if (RZ != NULL)
      {
        double sAZ, sBZ, sEZ, sre;
        float_image_t *EZ = pst_height_map_compare(CZ, RZ, EW, TRUE, &sAZ, &sBZ, &sEZ, &sre);
        if (writeImages)
          { float_image_mscale_write_file(EZ, filePrefix, levelQ, iterQ, "eZ", indent);
	    float_image_mscale_write_file(RZ, filePrefix, levelQ, iterQ, "rZ", indent);
	  }
        if (writeError)
          { char *fileName = float_image_mscale_file_name(filePrefix, levelQ, iterQ, "eZ", "txt");
            fprintf(stderr, "%*swriting %s ...", indent, "", fileName);
            FILE* wr = fopen(fileName, "wt");
            assert(wr != NULL); 
            fprintf
              ( wr, "%2d %6d %6d %9d %14.11f  %14.11f %14.11f  %14.11f %14.11f\n",
                level, NX, NY, iter, change, sAZ, sBZ, sEZ, sre
              );
            if (wr == stdout) { fflush(wr); } else { fclose(wr); }
            fprintf(stderr, "\n");
            free(fileName);
          }
      }
  }
